// -*- tab-width:2 ; indent-tabs-mode:nil -*-
//:: case FutureQueueApplication

/* To verify this example:

  vct --silver=silicon --check-history FutureQueue.java SpecifiedThread.java FutureWorker.java AbstractFutureQueue.java
  
 */
//begin(all)
class Sender extends Thread {

  AbstractFutureQueue queue;
  int input[];
  //@ seq<int> vals;
  
  /*@
    resource preFork(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.fut)
      ** Value(this.input) ** PointsTo(queue.fut_active,1/4,true)
      ** Value(this.vals)
      ** input.length==|vals|
      ** (\forall* int i; 0 <= i && i < |vals| ; Perm(input[i],1))
      ** (\forall  int i; 0 <= i && i < |vals| ; input[i]==vals[i])
      ** Future(queue.fut,1/2,queue.fut.put_all(vals))
      ** true;
      
    resource postJoin(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.fut)
      ** Value(this.input) ** PointsTo(queue.fut_active,1/4,true)
      ** Value(this.vals)
      ** input.length==|vals|
      ** (\forall* int i; 0 <= i && i < |vals| ; Perm(input[i],1))
      ** (\forall  int i; 0 <= i && i < |vals| ; input[i]==vals[i])
      ** Future(queue.fut,1/2,empty)
      ** true;
  @*/
  
  /*@
    given    seq<int>vals;
    requires input.length==|vals| ** queue!=null
      ** Value(queue.fut) ** Future(queue.fut,1/2,queue.fut.put_all(vals))
      ** PointsTo(queue.fut_active,1/4,true)
      ** (\forall* int i; 0 <= i && i < |vals| ; Perm(input[i],1))
      ** (\forall  int i; 0 <= i && i < |vals| ; input[i]==vals[i]);
    ensures  Value(this.queue) ** queue!=null
      ** this.queue==queue
      ** Value(this.input) ** this.input==input
      ** Value(this.vals) ** this.vals==vals ** preFork(1);
  @*/
  public Sender(AbstractFutureQueue queue,int[] input){
    this.queue=queue;
    this.input=input;
    //@ this.vals=vals;
    //@ fold this.preFork@Thread(1);
    //@ fold this.preFork@Sender(1);
  }
  
  public void run(){
    //@ unfold preFork@Sender(1);

    int N=input.length;
    int i=0;
    
    //@ assert drop(vals,i) == vals; // Not automatically found by prover.
    //@ assert Future(queue.fut,1/2,queue.fut.put_all(drop(vals,i)));
    
    /*@
      loop_invariant 0 <= i ** i <= N 
      ** Value(this.queue) ** queue!=null ** Value(queue.fut)
      ** Value(this.input) ** PointsTo(queue.fut_active,1/4,true)
      ** Value(this.vals)
      ** N==input.length ** N==|vals|
      ** (\forall* int k; 0 <= k < |vals| ; Perm(input[k],1))
      ** (\forall  int k; 0 <= k < |vals| ; input[k]==vals[k])
      ** Future(queue.fut,1/2,queue.fut.put_all(drop(vals,i)));
    @*/
    while(i<N){
      //@ assert Future(queue.fut,1/2,queue.fut.put_all(drop(vals,i)));
      //@ assert drop(vals,i)==seq<int>{vals[i]}+drop(vals,i+1);
      //@ assert Future(queue.fut,1/2,queue.fut.put_all(seq<int>{vals[i]}+drop(vals,i+1)));
      //@ assert Future(queue.fut,1/2,queue.fut.put(input[i])*queue.fut.put_all(drop(vals,i+1)));
      queue.put(input[i])
      /*@ with { p=1/2 ; P = queue.fut.put_all(drop(vals,i+1));} @*/;
      //@ assert Future(queue.fut,1/2,queue.fut.put_all(drop(vals,i+1)));
      int j = i + 1;
      //@ assert i+1 == j;
      //@ assert drop(vals,i+1) == drop(vals,j);
      //@ assert Future(queue.fut,1/2,queue.fut.put_all(drop(vals,j)));
      i=i+1;
      //@ assert drop(vals,i) == drop(vals,j);
      //@ assert Future(queue.fut,1/2,queue.fut.put_all(drop(vals,i)));
    }
    //@ assert Future(queue.fut,1/2,queue.fut.put_all(drop(vals,N)));
    //@ fold this.postJoin@Thread(1);
    //@ fold this.postJoin@Sender(1);

  }
  

}


class Receiver extends Thread {

  AbstractFutureQueue queue;
  int output[];
  //@ seq<int> vals;
  
  /*@
    resource preFork(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.fut)
      ** Value(this.output) ** PointsTo(queue.fut_active,1/4,true)
      ** Value(this.vals)
      ** output.length==|vals|
      ** (\forall* int i; 0 <= i && i < output.length ; Perm(output[i],1))
      ** Future(queue.fut,1/2,queue.fut.get_all(vals))
      ** true;
      
    resource postJoin(frac p)=p==write
      ** Value(this.queue) ** queue!=null ** Value(queue.fut)
      ** Value(this.output) ** PointsTo(queue.fut_active,1/4,true)
      ** Value(this.vals)
      ** output.length==|vals|
      ** (\forall* int i; 0 <= i < |vals| ; Perm(output[i],1))
      ** (\forall  int i; 0 <= i < |vals| ; output[i]==vals[i])
      ** Future(queue.fut,1/2,empty)
      ** true;
  @*/
  
  /*@
    given    seq<int>vals;
    requires output.length==|vals| ** queue!=null ** Value(queue.fut)
      ** Future(queue.fut,1/2,queue.fut.get_all(vals))
      ** PointsTo(queue.fut_active,1/4,true)
      ** (\forall* int i; 0 <= i < output.length ; Perm(output[i],1));
    ensures  Value(this.queue) ** queue!=null
      ** this.queue==queue ** Value(this.output)
      ** this.output==output
      ** Value(this.vals) ** this.vals==vals ** preFork(1);
  @*/
  public Receiver(AbstractFutureQueue queue,int[] output){
    this.queue=queue;
    this.output=output;
    //@ this.vals=vals;
    //@ fold this.preFork@Thread(1);
    //@ fold this.preFork@Receiver(1);
  }

// begin(receiver)  
  public void run(){
    //@ unfold preFork@Receiver(1); // skip(receiver)
    int N=output.length, i=0;
    
    //@ assert drop(vals,i) == vals;
    //@ assert Future(queue.fut,1/2,queue.fut.get_all(drop(vals,i)));

    /*@ loop_invariant Value(queue) ** Value(queue.fut)
    ** Value(output) ** Value(vals) ** 0 <= i <= N ** N==|vals|
    ** PointsTo(queue.fut_active,1/4,true) ** N==output.length
    ** (\forall* int k; 0 <= k < N ; Perm(output[k],1))
    ** (\forall int k; 0 <= k < i ; output[k]==vals[k])
    ** Future(queue.fut,1/2,queue.fut.get_all(drop(vals,i))); @*/
    while(i<N){
      //@ assert Future(queue.fut,1/2,queue.fut.get_all(drop(vals,i)));
      //@ assert drop(vals,i)==seq<int>{vals[i]}+drop(vals,i+1);
      //@ assert Future(queue.fut,1/2,queue.fut.get_all(seq<int>{vals[i]}+drop(vals,i+1)));
      //@ assert Future(queue.fut,1/2,queue.fut.get(vals[i])*queue.fut.get_all(drop(vals,i+1)));
      output[i]=queue.get()/*@ with {
          p=1/2 ; P = queue.fut.get_all(drop(vals,i+1)); val = vals[i];} @*/;
      i=i+1;
    }
    //@ fold this.postJoin@Thread(1); // skip(receiver)
    //@ fold this.postJoin@Receiver(1); // skip(receiver)
  }
// end(receiver)
}
//end(all)

