//:: cases MathCaller
//:: tools silicon
//:: verdict Pass

/* To verify this example:

  vct --silver=silicon --check-history dummy/Future.java dummy/mpi/{Status,Datatype,Request,Intracomm}.java dummy/MPIWrapper.java src/MathCaller.java

 */

import java.io.Serializable;

import mpi.Intracomm;
import mpi.Request;

public class MathCaller {
    
    public void main(Object sendbuf, Object recvbuf) {
        /*@
           Future fut = new Future();
           create fut, fut.callMethod(0, 1, sendbuf, recvbuf, 0, 1, 1, 2);
        @*/

        int a = 15;
        int b = 3;
        
        int res = divide(sendbuf, recvbuf, a, b)
        /*@ with {fut = fut; p = 1; P = empty; r1 = 1; r2 = 2;} @*/;
        
        //@ assert res == 5;
    }
    
    /*@
        given Future fut;
        given frac p;
        given process P;
        given int r1;
        given int r2;
        requires fut != null ** p != none;
        requires Future(fut, p, fut.send(0, 1, sendbuf, 0, r1) *
                                fut.irecv(1, 0, recvbuf, 1, r2) *
                                fut.wait(r2) *
                                P);
        requires b != 0 ** a % b == 0;
        ensures Future(fut, p, P);
        ensures \result * b == a;
    @*/
    public int divide(Object sendbuf, Object recvbuf, int a, int b) {
        //@ assert b != 0 ** a % b == 0;
        
        // Call call = new Call(1, "divide", new Serializable[]{a, b});
        // Object[] sendbuf = new Object[]{call};

        MPIWrapper mpi = new MPIWrapper();
        Intracomm COMM_WORLD = mpi.COMM_WORLD();

        //@ assume COMM_WORLD.Rank() == 0;
        
        COMM_WORLD.Send(sendbuf, 0, 1, mpi.OBJECT(), 1, 0)
        /*@ with {fut = fut; p = p; P = fut.irecv(1, 0, recvbuf, 1, r2) * fut.wait(r2) * P; r = r1;} @*/;        
        
        // Object[] recvbuf = new Object[1];
        Request req = COMM_WORLD.Irecv(recvbuf, 0, 1, mpi.OBJECT(), 1, 1)
        /*@ with {fut = fut; p = p; P = fut.wait(r2) * P; r = r2;} @*/;
        
        req.Wait()/*@ with {fut = fut; p = p; P = P;} @*/;

        // int res = (Integer) recvbuf[0];
        int res = 0;
        
        //@ assume res * b == a;
        return res;
        //@ refute false;
    }
}
