import java.io.Serializable;
import java.security.InvalidParameterException;
import java.util.Arrays;

public class MergeSortImpl extends ActiveObject {
    public MergeSortImpl(int objectId) {
        super(objectId);
    }

    public int[] sort(int[] list, int iteration) {
        int me = MPAO.getWorkerId();
        int workers = MPAO.getWorkerCount();

        if (list.length > 1) {
            int mid = (list.length > 10) ? list.length / 2 : list.length / 2;
            int workerId = me + (workers / (int) (Math.pow(2, iteration)));

            int[] left, right;
            if (workerId == me) {
                // Local sort right half
                right = Arrays.copyOfRange(list, mid, list.length);
                right = sort(right, iteration + 1);
                // Local sort left half
                left = Arrays.copyOfRange(list, 0, mid);
                left = sort(left, iteration + 1);
            } else {
                // Remote sort right half
                RemoteFuture<MergeSortRemote> msf = MPAO.getFactory(workerId).newMergeSort();
                right = Arrays.copyOfRange(list, mid, list.length);
                RemoteFuture<int[]> f = msf.get().sort(right, iteration + 1);

                // Local sort left half
                left = Arrays.copyOfRange(list, 0, mid);
                left = sort(left, iteration + 1);

                // Get right result
                right = f.get();
            }

            // working...
            int a1;
            int a2 = 0;

            for (int i = 0; i < (100000 * list.length); i++) {
                a1 = a2 + i;
                a2 = a1;
            }

            // MERGE

            int i = 0;
            int j = 0;

            while (i < left.length && j < right.length) {
                if (left[i] <= right[j]) {
                    list[i + j] = left[i];
                    i++;
                } else {
                    list[i + j] = right[j];
                    j++;
                }
            }
            while (i < left.length) {
                list[i + j] = left[i];
                i++;
            }
            while (j < right.length) {
                list[i + j] = right[j];
                j++;
            }
        }
        return list;
    }

    @Override
    public MergeSortRemote getRemote() {
        return new MergeSortRemote(MPAO.getWorkerId(), getObjectId());
    }

    @Override
    protected Serializable callMethod(Call call) {
        Serializable[] parameters = call.getParameters();
        switch (call.getSignature()) {
        case "sort":
            return this.sort((int[]) parameters[0], (Integer) parameters[1]);
        default:
            throw new InvalidParameterException();
        }
    }

}
