import java.io.Serializable;
import java.util.concurrent.atomic.AtomicBoolean;

//begin(all)
public class RemoteFuture<T extends Serializable> {
    /*@
        public resource getToken() = Perm(this.dummy, 1);
        // TODO Make postcondition
        public resource postGet(T value) = valid(source?, dest?, \result, tag?);

        private boolean dummy;
        private frac getExecuted;
        // begin(invariant)
        resource csl_invariant() = Value(this.done) ** Perm(this.done.val, 1/2)
         ** Perm(this.getExecuted, 1)
         ** (this.getExecuted != none ==> Perm(this.dummy, this.getExecuted))
         ** (!this.done.val ==> this.getExecuted == none)
         ** (this.done.val ==> (Value(this.done.val)
             ** Value(this.value)
             ** [1 - this.getExecuted]postGet(this.value)))
         ** Value(this.waiting) ** Perm(this.waiting.val, 1/2)
         ** (!this.waiting.val ==> Perm(this.waiting.val, 1/2)
             ** PointsTo(this.done.val, 1/2, false)
             ** Value(result) ** result.getToken()
             ** Perm(this.value, 1))
         ** (this.waiting.val ==> Value(this.waiting.val));
        // end(invariant)
    @*/

    private AtomicBoolean done;
    private AtomicBoolean waiting;

    private MPAOResult result;
    private T value;

    /*@
        requires result.getToken();
        ensures  getToken();
    @*/
    public RemoteFuture(MPAOResult result) {
        this.done = new AtomicBoolean();
        this.waiting = new AtomicBoolean();

        this.result = result;

        //@ this.getExecuted = none;
        //@ fold this.getToken();
        //@ refute false; //skip(all)
    }

    /*@
        given frac p;
        requires p > none ** [p]getToken();
        ensures  [p]postGet(\result);
      @*/
    public T get() {
        boolean d = false;

        while (!waiting.get()) {
            if (waiting.compareAndSet(false, true)) {
                // TODO Waar toveren we hier de future vandaan?
                this.value = (T) this.result.get()
                /*@ with {fut = fut; p = p; P(msg) = P(msg);} @*/;
                //@ fold this.postGet(this.value);
                done.set(true);
            }
        }

        /*@
            loop_invariant d ==> Value(this.done) ** Value(this.done.val) **
                                 this.done.val ** Value(this.value);
        @*/
        while(!d) {
            d = this.done.get();
        }

        //@ unfold [p]this.getToken();
        /*@
          atomic(this) {
            frac oldGetExecuted;
            oldGetExecuted = this.getExecuted;
            assert oldGetExecuted + p <= write;
            this.getExecuted = oldGetExecuted + p;
          }
        @*/

        return this.value;
        //@ refute false; //skip(all)
    }
}
//end(all)
